//------------------------------------------------------------------------------------------------
//
//   Project Blue - Pneumatic Tube Part
//
//------------------------------------------------------------------------------------------------

package gcewing.projectblue;

import java.util.*;
import static java.lang.Math.*;

import net.minecraft.client.renderer.entity.*;
import net.minecraft.client.renderer.RenderBlocks;
import net.minecraft.entity.item.*;
import net.minecraft.entity.player.*;
import net.minecraft.init.*;
import net.minecraft.inventory.*;
import net.minecraft.item.*;
import net.minecraft.nbt.*;
import net.minecraft.tileentity.*;
import net.minecraft.util.*;
import net.minecraft.world.*;
import net.minecraftforge.common.util.*;
import cpw.mods.fml.relauncher.*;

import codechicken.microblock.*;
import codechicken.multipart.*;
import codechicken.multipart.MultiPartRegistry.IPartFactory;
import codechicken.lib.data.*;
import codechicken.lib.lighting.*;
import codechicken.lib.raytracer.*;
import codechicken.lib.vec.*;

import mrtjp.projectred.core.ItemScrewdriver;
import mrtjp.projectred.transmission.*;

public class PneumaticTubePart extends JCenterPart implements ISidedHollowConnect {

	final static int tubeWidth = 6; // sixteenths
	final static double h = 0.5 * (tubeWidth / 16.0);
	static Cuboid6 centerBox = new Cuboid6(0.5 - h, 0.5 - h, 0.5 - h, 0.5 + h, 0.5 + h, 0.5 + h);
	static Cuboid6[] sideBoxes = new Cuboid6[6];
	static {
		for (int i = 0; i < 6; i++) {
			Transformation t = Rotation.sideRotations[i].at(codechicken.lib.vec.Vector3.center);
			sideBoxes[i] = new Cuboid6(0.5 - h, 0, 0.5 - h, 0.5 + h, 0.5 - h, 0.5 + h).apply(t);
		}
	}
	static Random random = new Random();
	
	final static double speed = 1/8d; //0.25;

	public int connections = 0;
	public byte outputSide = -1;
	public Payload[/*side*/] payloads = new Payload[6];
	public byte[/*side*/] tags = {-1, -1, -1, -1, -1, -1};
		
	public boolean isConnectedOnSide(int i) {
		return (connections & (1 << i)) != 0;
	}
	
	@Override
	public String getType() {
		return "pb_pneumatictube";
	}
	
	@Override
	public int getHollowSize(int side) {
		return tubeWidth;
	}
	
	@Override
	public Iterable<Cuboid6> getCollisionBoxes() {
		List<Cuboid6> result = new ArrayList<Cuboid6>();
		result.add(centerBox);
		for (int i = 0; i < 6; i++)
			if (isConnectedOnSide(i))
				result.add(sideBoxes[i]);
		return result;
	}

	@Override
	public Iterable<IndexedCuboid6> getSubParts() {
		List<IndexedCuboid6> result = new ArrayList<IndexedCuboid6>();
		result.add(new IndexedCuboid6(6, centerBox));
		for (int i = 0; i < 6; i++)
			if (isConnectedOnSide(i))
				result.add(new IndexedCuboid6(i, sideBoxes[i]));
		return result;
	}

	@Override
	public boolean activate(EntityPlayer player, MovingObjectPosition hit, ItemStack stack) {
		if (stack != null) {
			Item item = stack.getItem();
			//if (!world().isRemote)
			//	System.out.printf("PneumaticTubePart.activate: with %s[%s]\n",
			//		item.getClass().getName(), stack.getItemDamage());
			if (item instanceof SprayCanItem) {
				useSprayCan(hit, stack);
				return true;
			}
			if (item instanceof ItemScrewdriver) {
				useScrewdriver(hit, stack);
				return true;
			}
			if (item instanceof ItemPartWire) {
				useWire(hit, stack);
				return true;
			}
		}
		return false;
	}
	
	void useSprayCan(MovingObjectPosition hit, ItemStack stack) {
		if (!world().isRemote) {
			int side = hitSide(hit);
			SprayCanItem can = (SprayCanItem)stack.getItem();
			int tag = can.getColor(stack);
			if (tag > 15)
				tag = -1;
			tags[side] = (byte)tag;
			outputSide = -1;
			stack.setItemDamage(stack.getItemDamage() + 1);
			setupChanged();
		}
	}
	
	void useScrewdriver(MovingObjectPosition hit, ItemStack stack) {
		if (!world().isRemote) {
			if (hit.subHit < 6) {
				int side = hit.subHit;
				if (outputSide == side)
					outputSide = -1;
				else
					outputSide = (byte)side;
				stack.setItemDamage(stack.getItemDamage() + 1);
				setupChanged();
			}
		}
	}
	
	void useWire(MovingObjectPosition hit, ItemStack stack) {
		if (!world().isRemote && hit.subHit == 6) {
			int meta = stack.getItemDamage();
			WireDef.WireDef def = (WireDef.WireDef)WireDef.values().apply(meta);
			String type = "pb_" + def.wireType();
			WirePart wire = (WirePart)MultiPartRegistry.createPart(type, false);
			if (wire != null) {
				wire.preparePlacement(hit.sideHit ^ 1, meta);
				tile().addPart(world(), new BlockCoord(x(), y(), z()), wire);
			}
		}
	}
	
	int hitSide(MovingObjectPosition hit) {
		int side = hit.subHit;
		if (side >= 6)
			side = hit.sideHit;
		return side;
	}
		
	public void sendSetupUpdate() {
		MCDataOutput data = getWriteStream();
		writeSetup(data);
		data.writeByte(-1);
	}
	
	public void sendPayloadUpdate() {
		MCDataOutput data = getWriteStream();
		writePayloads(data);
		data.writeByte(0);
	}
	
	@Override
	public void read(MCDataInput data) {
		readDesc(data);
	}
	
	@Override
	public void writeDesc(MCDataOutput data) {
		writeSetup(data);
		writePayloads(data);
		data.writeByte(0);
	}
	
	@Override
	public void readDesc(MCDataInput data) {
		for (;;) {
			int tag = data.readByte();
			//System.out.printf("PneumaticTubePart.readDesc: tag %s\n", tag);
			switch (tag) {
				case 0:
					return;
				case -1:
					tile().markRender();
					return;
				case 1:
					readSetup(data);
					break;
				case 2:
					readPayloads(data);
					break;
				default:
					throw new RuntimeException(
						String.format("PneumaticTubePart.read() got invalid tag byte %s\n", tag));
			}
		}
	}

	protected void writeSetup(MCDataOutput data) {
		data.writeByte(1);
		data.writeByte(connections & 0xff);
		data.writeByte(outputSide);
		for (int i = 0; i < 6; i++)
			data.writeByte(tags[i]);
	}
	
	protected void readSetup(MCDataInput data) {
		connections = data.readByte();
		outputSide = data.readByte();
		for (int i = 0; i < 6; i++)
			tags[i] = data.readByte();
	}
	
	protected void writePayloads(MCDataOutput data) {
		data.writeByte(2);
		int mask = 0;
		for (int i = 0; i < 6; i++)
			if (payloads[i] != null)
				mask |= (1 << i);
		data.writeByte((byte)mask);
		for (int i = 0; i < 6; i++)
			if (payloads[i] != null)
				payloads[i].writeDesc(data);
	}
	
	protected void readPayloads(MCDataInput data) {
		int mask = data.readByte();
		for (int i = 0; i < 6; i++) {
			if ((mask & (1 << i)) == 0)
				payloads[i] = null;
			else {
				Payload payload = payloads[i];
				if (payload == null) {
					payload = new Payload();
					payloads[i] = payload;
				}
				payload.readDesc(data);
				System.out.printf("PneumaticTubePart.readPayloads: pos %.3f",
					payload.position);
				if (tile() != null)
					System.out.printf(" for tube at (%s,%s,%s)", x(), y(), z());
				System.out.printf("\n");
			}
		}
	}

	@Override
	public void save(NBTTagCompound nbt) {
		nbt.setInteger("connections", connections);
		nbt.setInteger("output", outputSide + 1);
		nbt.setByteArray("tags", tags);
		NBTTagList payList = new NBTTagList();
		for (int i = 0; i < 6; i++)
			if (payloads[i] != null) {
				NBTTagCompound pnbt = new NBTTagCompound();
				pnbt.setInteger("leg", i);
				payloads[i].save(pnbt);
				payList.appendTag(pnbt);
			}
		if (payList.tagCount() > 0)
			nbt.setTag("payloads", payList);
	}

	@Override
	public void load(NBTTagCompound nbt) {
		setConnections(nbt.getInteger("connections"));
		outputSide = (byte)(nbt.getInteger("output") - 1);
		byte[] tags = nbt.getByteArray("tags");
		if (tags.length >= 6)
			this.tags = tags;
		NBTTagList payList = nbt.getTagList("payloads", Constants.NBT.TAG_COMPOUND);
		for (int i = 0; i < payList.tagCount(); i++) {
			NBTTagCompound pnbt = payList.getCompoundTagAt(i);
			int side = pnbt.getInteger("leg");
			Payload pay = new Payload();
			pay.load(pnbt);
			network.addPayload(pay, this, side);
		}
	}

	@Override
	public boolean renderStatic(codechicken.lib.vec.Vector3 pos, int pass) {
		if (pass == 0)
			PneumaticTubeRenderer.renderStaticInWorld(this, pos.x, pos.y, pos.z);
		return false;
	}
	
	@Override
	public void renderDynamic(codechicken.lib.vec.Vector3 pos, float frame, int pass) {
		if (pass == 0)
			PneumaticTubeRenderer.renderDynamicInWorld(this, pos.x, pos.y, pos.z, frame);
	}
	
//	@Override
//	public void drawBreaking(RenderBlocks rb) {
//	}

	@Override
	public void onAdded() {
		super.onAdded();
		if (!world().isRemote)
			updateConnections();
	}
	
	@Override
	public void onRemoved() {
		if (!world().isRemote) {
			connections = 0;
			dropUncontainedPayloads();
		}
		super.onRemoved();
	}

	@Override
	public void onPartChanged(TMultiPart part) {
		if (!world().isRemote)
			updateConnections();
	}
	
	@Override
	public void onNeighborChanged() {
		if (!world().isRemote)
			updateConnections();
	}

	void updateConnections() {
		int mask = 0;
		for (int i = 0; i < 6; i++) {
			ForgeDirection d = ForgeDirection.getOrientation(i);
			TileEntity te = world().getTileEntity(x() + d.offsetX, y() + d.offsetY, z() + d.offsetZ);
			if (canConnectToTube(te, i ^ 1)) {
				if (tubeCanConnect(this, i))
					mask |= 1 << i;
			}
		}
		if (connections != mask) {
			setConnections(mask);
			dropUncontainedPayloads();
			setupChanged();
		}
	}
	
	void setConnections(int mask) {
		connections = mask;
	}
	
	void dropUncontainedPayloads() {
		for (int i = 0; i < 6; i++) {
			if ((connections & (1 << i)) == 0) {
				Payload pay = payloads[i];
				if (pay != null) {
					//payloads[i] = null;
					ForgeDirection d = ForgeDirection.getOrientation(i);
					double p = 1 - pay.position;
					double v = -pay.velocity * 0.5;
					EntityItem entity = new EntityItem(world(),
						x() + 0.5 * (1 + d.offsetX * p), 
						y() + 0.5 * (1 + d.offsetY * p), 
						z() + 0.5 * (1 + d.offsetZ * p),
						pay.stack);
					double r = 0.05;
					entity.motionX = v * d.offsetX + random.nextGaussian() * r;
					entity.motionY = v * d.offsetY + random.nextGaussian() * r;
					entity.motionZ = v * d.offsetZ + random.nextGaussian() * r;
					entity.hoverStart = 0;
					world().spawnEntityInWorld(entity);
					network.removePayload(pay);
				}
			}
		}
	}

	static boolean canConnectToTube(TileEntity te, int side) {
		if (te instanceof ISidedInventory)
			return ((ISidedInventory)te).getAccessibleSlotsFromSide(side).length > 0;
		if (te instanceof IInventory)
			return ((IInventory)te).getSizeInventory() > 0;
		if (te instanceof TileMultipart) {
			TileMultipart tmp = (TileMultipart)te;
			TMultiPart part = tmp.partMap(6);
			if (part instanceof PneumaticTubePart)
				return tubeCanConnect(part, side);
		}
		return false;
	}
	
	static boolean tubeCanConnect(TMultiPart tube, int side) {
		TMultiPart part = tube.tile().partMap(side);
		if (part == null)
			return true;
		if (part instanceof TFacePart) {
			TFacePart f = (TFacePart)part;
			return (f.redstoneConductionMap() & (1 << side)) != 0;
		}
		return false;
	}
	
	void markDirty() {
		TileEntity te  = (TileEntity)tile();
		System.out.printf("PneumaticTubePart.markDirty: tile = %s\n", te);
		if (te != null)
			te.markDirty();
	}
	
	@Override
	public boolean doesTick() {
		return true;
	}
	
//	@Override
//	public void update() {
//		if (world().isRemote)
//			clientUpdate();
//		else
//			serverUpdate();
//	}
//	
//	void clientUpdate() {
//		for (int i = 0; i < 6; i++) {
//			Payload pay = payloads[i];
//			if (pay != null) {
//				double newPos = pay.position + pay.velocity;
//				if (newPos >= 0.0 && newPos <= 1.0)
//					pay.position = newPos;
//			}
//		}
//	}
//				
//	void serverUpdate() {
//		for (int i = 0; i < 6; i++) {
//			Payload pay = payloads[i];
//			if (pay != null)
//				updatePayload(pay);
//		}
//	}
	
	void updatePayload(Payload pay) {
		int i = pay.side;
		double newPos = pay.position + pay.velocity;
		if (newPos > 1.0) {
			//pay.position = 2.0 - newPos;
			//pay.velocity = -pay.velocity;
			System.out.printf("PneumaticTubePart.serverUpdate: route internal side %s pos %.3f vel %.3f\n",
				i, pay.position, pay.velocity);
			routeInternal(pay, i, 2.0 - newPos);
		}
		else if (newPos < 0.0) {
			//pay.position = -newPos;
			//pay.velocity = -pay.velocity;
			System.out.printf("PneumaticTubePart.serverUpdate: route external side %s pos %.3f vel %.3f\n",
				i, pay.position, pay.velocity);
			routeExternal(pay, i, -newPos);
		}
		else {
			System.out.printf("PneumaticTubePart.serverUpdate: move internal side %s pos %.3f vel %.3f\n",
				i, pay.position, pay.velocity);
			//pay.position = newPos;
			//payloadChanged();
			network.movePayloadToPosition(pay, newPos);
		}
	}
	
	void routeInternal(Payload pay, int incomingSide, double newPosition) {
		int[] candidateRoutes = new int[6];
		int numCandidateRoutes = 0;
		// If an explicit route is specified, use it
		if (outputSide != -1 && outputSide != incomingSide && isConnectedOnSide(outputSide))
			candidateRoutes[numCandidateRoutes++] = outputSide;
		else {
			// Otherwise, route by tag.
			// If a leg is tagged, it will only accept a payload with a matching tag.
			// If a leg is untagged, it will accept untagged payloads and tagged payloads
			// for which no other outgoing leg has a matching tag.
			boolean taggedRouteExists = false;
			//boolean openRouteExists = false;
			if (pay.tag >= 0)
				for (int j = 0; j < 6; j++)
					if (incomingSide != j && tags[j] == pay.tag) {
						taggedRouteExists = true;
						break;
					}
			// Search outgoing legs for an acceptable route.
			for (int j = 1; j <= 5; j++) {
				int k = (incomingSide + j) % 6;
				if (isConnectedOnSide(k))
					if ((tags[k] >= 0) ? (tags[k] == pay.tag) : (pay.tag == 0 || !taggedRouteExists))
						candidateRoutes[numCandidateRoutes++] = k;
			}
		}
		// Search candidate routes for one that is uncongested
		for (int i = 0; i < numCandidateRoutes; i++) {
			int k = candidateRoutes[i];
			if (payloads[k] == null) {
				network.movePayloadToSide(pay, k, newPosition, -pay.velocity);
				return;
			}
			//else if (movingInSameDirection(pay, payloads[k]))
			//	openRouteExists = true;
		}
		//if (!openRouteExists)
		//	bounce(pay);
	}
	
	void routeExternal(Payload pay, int i, double newPosition) {
		TileEntity te = getAdjacentTileEntity(i);
		if (te instanceof ISidedInventory)
			if (routeToSidedInventory(pay, i, (ISidedInventory)te))
				return;
		if (te instanceof IInventory)
			if (routeToInventory(pay, i, (IInventory)te))
				return;
		PneumaticTubePart other = getPneumaticTubePart(te);
		if (other != null) {
			if (routeToAdjacentTube(pay, i, other, newPosition))
				return;
		}
		//bounce(pay);
	}
	
	boolean routeToAdjacentTube(Payload pay, int i, PneumaticTubePart other, double newPosition) {
		int j = i ^ 1;
		if (other.payloads[j] == null) {
//			pay.tube = other;
//			other.payloads[j] = pay;
//			pay.side = j;
			network.movePayloadToTubeAndSide(pay, other, j, newPosition, -pay.velocity);
			return true;
		}
		else
			return movingInSameDirection(pay, other.payloads[j]);
	}
	
	boolean routeToSidedInventory(Payload pay, int i, ISidedInventory te) {
		int side = i ^ 1;
		int[] slots = te.getAccessibleSlotsFromSide(side);
		return routeToInventorySlots(pay, side, te, slots.length, slots);
	}

	boolean routeToInventory(Payload pay, int i, IInventory te) {
		return routeToInventorySlots(pay, i ^ 1, te, te.getSizeInventory(), null);
	}
	
	boolean routeToInventorySlots(Payload pay, int side, IInventory te, int numSlots, int[] slots) {
		boolean canInsert = false;
		ItemStack stack = pay.stack;
		int maxStack = te.getInventoryStackLimit();
		if (stack.isStackable()) {
			for (int i = 0; i < numSlots && stack.stackSize > 0; i++) {
				int slot = (slots == null) ? i : slots[i];
				ItemStack slotStack = te.getStackInSlot(slot);
				if (slotStack != null && stack.isItemEqual(slotStack)) {
					canInsert = true;
					int n = min(maxStack, slotStack.getMaxStackSize()) - slotStack.stackSize;
					n = min(n, stack.stackSize);
					if (n > 0) {
						slotStack.stackSize += n;
						stack.stackSize -= n;
						te.setInventorySlotContents(slot, slotStack);
					}
				}
			}
		}
		if (stack.stackSize > 0) {
			for (int i = 0; i < numSlots; i++) {
				int slot = (slots == null) ? i : slots[i];
				if (canInsertItem(te, slot, stack, side)) {
					canInsert = true;
					if (te.getStackInSlot(slot) == null) {
						te.setInventorySlotContents(slot, stack);
						stack = null;
						break;
					}
				}
			}
		}
		if (stack == null || stack.stackSize == 0)
			network.removePayload(pay);
		return canInsert;
	}
	
	boolean canInsertItem(IInventory te, int slot, ItemStack stack, int side) {
		if (te instanceof ISidedInventory)
			return ((ISidedInventory)te).canInsertItem(slot, stack, side);
		else
			return true;
	}
	
	boolean movingInSameDirection(Payload pay1, Payload pay2) {
		return signum(pay1.velocity) != signum(pay2.velocity);
	}
	
//	void payloadLeftTube(int i) {
////		payloads[i] = null;
//		payloadChanged();
//	}
	
//	void bounce(Payload pay) {
//		pay.velocity = -pay.velocity;
//		payloadChanged();
//	}
	
	public void setupChanged() {
		markDirty();
		sendSetupUpdate();
	}
		
//	public void payloadChanged() {
//		markDirty();
//		schedulePayloadUpdate();
//	}
		
	TileEntity getAdjacentTileEntity(int side) {
		ForgeDirection d = ForgeDirection.getOrientation(side);
		return world().getTileEntity(x() + d.offsetX, y() + d.offsetY, z() + d.offsetZ);
	}
	
	PneumaticTubePart getPneumaticTubePart(TileEntity te) {
		if (te instanceof TileMultipart) {
			TileMultipart tmp = (TileMultipart)te;
			TMultiPart part = tmp.partMap(6);
			if (part instanceof PneumaticTubePart)
				return (PneumaticTubePart)part;
		}
		return null;
	}
	
	public boolean willAcceptFromSide(int side) {
		return payloads[side] == null;
	}
	
	public void insertFromSide(int side, ItemStack stack, int tag) {
		System.out.printf("PneumaticTubePart.insertFromSide %s: %s with tag %s\n", side, stack, tag);
		Payload pay = new Payload(stack, tag);
		network.insertPayload(pay, this, side, 0, speed);
	}
	
	@Override
	public void onWorldSeparate() {
		System.out.printf("PneumaticTubePart.onWorldSeparate: at (%s,%s,%s)\n", x(), y(), z());
		for (Payload pay : payloads)
			if (pay != null)
				network.forgetPayload(pay);
		scheduledPayloadUpdates.remove(this);
	}
	
//------------------------------------------------------------------------------------------------

	protected static Set<PneumaticTubePart> scheduledPayloadUpdates =
		new HashSet<PneumaticTubePart>();
	
	public void schedulePayloadUpdate() {
		scheduledPayloadUpdates.add(this);
	}
	
	public static void sendScheduledUpdates() {
		for (PneumaticTubePart part : scheduledPayloadUpdates) {
			if (part.tile() != null) {
				System.out.printf("PneumaticTubePart.sendScheduledUpdates: for tube at (%s,%s,%s)\n",
					part.x(), part.y(), part.z());
				part.sendPayloadUpdate();
			}
		}
		scheduledPayloadUpdates.clear();
	}
	
	public static void onServerTickEnd() {
		network.serverUpdate();
		sendScheduledUpdates();
	}

//------------------------------------------------------------------------------------------------

	public static class Factory implements IPartFactory {
	
		public TMultiPart createPart(String name, boolean client) {
			return new PneumaticTubePart();
		}
	
	}

//------------------------------------------------------------------------------------------------

	public static class Payload {
		public PneumaticTubePart tube;
		public int side;
		public ItemStack stack;
		public int tag;
		public double position; // 0 at face, 1 at centre
		public double velocity; // positive towards centre
	
//		public Payload(PneumaticTubePart tube, int side) {
//			setTubeAndSide(tube, side);
//			//this.tube = tube;
//			//this.side = side;
//			onAdded();
//		}
//		
//		public Payload(PneumaticTubePart tube, int side, ItemStack stack, int tag, double position, double velocity) {
//			this(tube, side);
//			this.stack = stack;
//			this.tag = tag;
//			this.position = position;
//			this.velocity = velocity;
//		}

		public Payload() {
		}
			
		public Payload(ItemStack stack, int tag) {
			this.stack = stack;
			this.tag = tag;
		}
		
//		protected abstract void onAdded();
//		protected abstract void onRemoved();
		
		public void writeDesc(MCDataOutput data) {
			data.writeItemStack(stack);
			data.writeInt(tag);
			data.writeFloat((float)position);
			data.writeFloat((float)velocity);
		}
		
		public void readDesc(MCDataInput data) {
			stack = data.readItemStack();
			tag = data.readInt();
			position = data.readFloat();
			velocity = data.readFloat();
		}
		
		public void save(NBTTagCompound nbt) {
			stack.writeToNBT(nbt);
			nbt.setInteger("tag", tag + 1);
			nbt.setDouble("position", position);
			nbt.setDouble("velocity", velocity);
		}
		
		public void load(NBTTagCompound nbt) {
			stack = ItemStack.loadItemStackFromNBT(nbt);
			tag = nbt.getInteger("tag") - 1;
			position = nbt.getDouble("position");
			velocity = nbt.getDouble("velocity");
		}
		
		public void setTubeAndSide(PneumaticTubePart tube, int side) {
			if (this.tube != null)
				this.tube.payloads[this.side] = null;
			this.tube = tube;
			this.side = side;
			if (tube != null)
				tube.payloads[side] = this;
		}
		
		public void setLocationAndVelocity(PneumaticTubePart tube, int side, double position, double velocity) {
			setTubeAndSide(tube, side);
			setPositionAndVelocity(position, velocity);
		}
		
		public void setPosition(double position) {
			this.position = position;
		}
		
		public void setPositionAndVelocity(double position, double velocity) {
			this.position = position;
			this.velocity = velocity;
		}
		
	}
	
//------------------------------------------------------------------------------------------------

	public static Network network = new Network();
		
	public static class Network {
	
		Set<Payload> payloads = new HashSet<Payload>();
		List<Payload> deadPayloads = new ArrayList<Payload>();
		
		public void serverUpdate() {
			for (Payload pay : deadPayloads)
				payloads.remove(pay);
			deadPayloads.clear();
			for (Payload pay : payloads) {
				if (pay.tube != null)
					pay.tube.updatePayload(pay);
			}
		}
		
		public void addPayload(Payload pay, PneumaticTubePart tube, int side) {
			pay.setTubeAndSide(tube, side);
			payloads.add(pay);
			tube.markDirty();
			tube.schedulePayloadUpdate();
		}
		
		public void insertPayload(Payload pay, PneumaticTubePart tube, int side, double position, double velocity) {
			addPayload(pay, tube, side);
			pay.setPositionAndVelocity(position, velocity);
			tube.markDirty();
			tube.schedulePayloadUpdate();
		}
		
		public void movePayloadToPosition(Payload pay, double position) {
			pay.setPosition(position);
			pay.tube.markDirty();
			pay.tube.schedulePayloadUpdate();
		}

		public void movePayloadToSide(Payload pay, int side, double position, double velocity) {
			pay.setLocationAndVelocity(pay.tube, side, position, velocity);
			pay.tube.markDirty();
			pay.tube.schedulePayloadUpdate();
		}
		
		public void movePayloadToTubeAndSide(Payload pay, PneumaticTubePart tube, int side, double position, double velocity) {
			pay.tube.markDirty();
			pay.tube.schedulePayloadUpdate();
			pay.setLocationAndVelocity(tube, side, position, velocity);
			tube.markDirty();
			tube.schedulePayloadUpdate();
		}
		
		public void removePayload(Payload pay) {
			PneumaticTubePart tube = pay.tube;
			pay.setTubeAndSide(null, 0);
			forgetPayload(pay);
			tube.markDirty();
			tube.schedulePayloadUpdate();
		}
		
		public void forgetPayload(Payload pay) {
			deadPayloads.add(pay);
		}
	
	}

//------------------------------------------------------------------------------------------------

	public static class WireFactory implements IPartFactory {
	
		public TMultiPart createPart(String name, boolean client) {
			if (name.equals("pb_pr_insulated"))
				return new InsulatedTubeWire();
			return null;
		}
		
	}

//------------------------------------------------------------------------------------------------

	public static class InsulatedTubeWire extends InsulatedRedAlloyPart {
	
		@Override
		public String getType() {
			return "pb_pr_insulated";
		}
	
		@Override
		public boolean canStay() {
			return true;
		}
	
		@SideOnly(Side.CLIENT)
		@Override 
		public void doBreakTessellation(RenderBlocks renderBlocks) {
			RenderTubeWire.renderBreakingOverlay(renderBlocks.overrideBlockTexture, this);
		}
		
		@SideOnly(Side.CLIENT)
		@Override
		public void doDynamicTessellation(codechicken.lib.vec.Vector3 pos, float frame, int pass) {
			//RenderTubeWire.render(this, pos);
			PneumaticTubeWireRenderer.render(this, pos.x, pos.y, pos.z, new PBStaticRenderer(this));
		}
	
		@SideOnly(Side.CLIENT)
		@Override
		public void doStaticTessellation(codechicken.lib.vec.Vector3 pos, int pass) {
			//RenderTubeWire.render(this, pos);
			PneumaticTubeWireRenderer.render(this, pos.x, pos.y, pos.z, new PBStaticRenderer(this));
		}

	}

//------------------------------------------------------------------------------------------------

}
